package com.sun.xml.internal.ws.api.message;

import java.util.List;
import com.sun.xml.internal.ws.binding.SOAPBindingImpl;
import java.util.HashSet;
import com.sun.xml.internal.ws.api.WSBinding;
import com.sun.xml.internal.ws.api.model.wsdl.WSDLPort;
import com.sun.xml.internal.ws.api.addressing.WSEndpointReference;
import com.sun.xml.internal.ws.api.addressing.AddressingVersion;
import java.util.NoSuchElementException;
import com.sun.istack.internal.Nullable;
import com.sun.istack.internal.NotNull;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Set;
import javax.xml.namespace.QName;
import com.sun.xml.internal.ws.api.SOAPVersion;
import java.util.BitSet;
import java.util.ArrayList;

public class HeaderList extends ArrayList<Header> implements MessageHeaders
{
    private static final long serialVersionUID = -6358045781349627237L;
    private int understoodBits;
    private BitSet moreUnderstoodBits;
    private SOAPVersion soapVersion;
    
    @Deprecated
    public HeaderList() {
        this.moreUnderstoodBits = null;
    }
    
    public HeaderList(SOAPVersion soapVersion) {
        this.moreUnderstoodBits = null;
        this.soapVersion = soapVersion;
    }
    
    public HeaderList(HeaderList that) {
        super(that);
        this.moreUnderstoodBits = null;
        this.understoodBits = that.understoodBits;
        if (that.moreUnderstoodBits != null) {
            this.moreUnderstoodBits = (BitSet)that.moreUnderstoodBits.clone();
        }
    }
    
    public HeaderList(MessageHeaders that) {
        super(that.asList());
        this.moreUnderstoodBits = null;
        if ((that instanceof HeaderList)) {
            HeaderList hThat = (HeaderList)that;
            this.understoodBits = hThat.understoodBits;
            if (hThat.moreUnderstoodBits != null) {
                this.moreUnderstoodBits = (BitSet)hThat.moreUnderstoodBits.clone();
            }
        }
        else {
            Set<QName> understood = that.getUnderstoodHeaders();
            if (understood != null) {
                for (QName qname : understood) {
                    this.understood(qname);
                }
            }
        }
    }
    
    @Override
    public int size() {
        return super.size();
    }
    
    @Override
    public boolean hasHeaders() {
        return !this.isEmpty();
    }
    
    @Deprecated
    public void addAll(Header... headers) {
        super.addAll(Arrays.asList(headers));
    }
    
    @Override
    public Header get(int index) {
        return (Header)super.get(index);
    }
    
    public void understood(int index) {
        if (index >= this.size()) {
            throw new ArrayIndexOutOfBoundsException(index);
        }
        if (index < 32) {
            this.understoodBits |= 1 << index;
        }
        else {
            if (this.moreUnderstoodBits == null) {
                this.moreUnderstoodBits = new BitSet();
            }
            this.moreUnderstoodBits.set(index - 32);
        }
        return;
    }
    
    public boolean isUnderstood(int index) {
        if (index >= this.size()) {
            throw new ArrayIndexOutOfBoundsException(index);
        }
        if (index < 32) {
            return this.understoodBits == (this.understoodBits | 1 << index);
        }
        return this.moreUnderstoodBits != null && this.moreUnderstoodBits.get(index - 32);
    }
    
    @Override
    @Deprecated
    public void understood(@NotNull Header header) {
        int sz = this.size();
        int i = 0;
        while (i < sz) {
            if (this.get(i) == header) {
                this.understood(i);
                return;
            }
            ++i;
        }
        throw new IllegalArgumentException();
    }
    
    @Nullable
    @Override
    public Header get(@NotNull String nsUri, @NotNull String localName, boolean markAsUnderstood) {
        int len = this.size();
        int i = 0;
        while (i < len) {
            Header h = this.get(i);
            if (h.getLocalPart().equals(localName) && h.getNamespaceURI().equals(nsUri)) {
                if (markAsUnderstood) {
                    this.understood(i);
                    return h;
                }
                return h;
            }
            else {
                ++i;
            }
        }
        return null;
    }
    
    @Deprecated
    public Header get(String nsUri, String localName) {
        return this.get(nsUri, localName, true);
    }
    
    @Nullable
    @Override
    public Header get(@NotNull QName name, boolean markAsUnderstood) {
        return this.get(name.getNamespaceURI(), name.getLocalPart(), markAsUnderstood);
    }
    
    @Nullable
    @Deprecated
    public Header get(@NotNull QName name) {
        return this.get(name, true);
    }
    
    @Deprecated
    public Iterator<Header> getHeaders(String nsUri, String localName) {
        return this.getHeaders(nsUri, localName, true);
    }
    
    @NotNull
    @Override
    public Iterator<Header> getHeaders(@NotNull final String nsUri, @NotNull final String localName, final boolean markAsUnderstood) {
        return new Iterator<Header>() {
            int idx = 0;
            Header next;
            
            @Override
            public boolean hasNext() {
                if (this.next == null) {
                    this.fetch();
                }
                return this.next != null;
            }
            
            @Override
            public Header next() {
                if (this.next == null) {
                    this.fetch();
                    if (this.next == null) {
                        throw new NoSuchElementException();
                    }
                }
                if (markAsUnderstood) {
                    assert HeaderList.this.get(this.idx - 1) == this.next;
                    HeaderList.this.understood(this.idx - 1);
                }
                Header r = this.next;
                this.next = null;
                return r;
            }
            
            private void fetch() {
                while (this.idx < HeaderList.this.size()) {
                    Header h = HeaderList.this.get(this.idx++);
                    if (h.getLocalPart().equals(localName) && h.getNamespaceURI().equals(nsUri)) {
                        this.next = h;
                        break;
                    }
                    continue;
                }
            }
            
            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }
    
    @NotNull
    @Override
    public Iterator<Header> getHeaders(@NotNull QName headerName, boolean markAsUnderstood) {
        return this.getHeaders(headerName.getNamespaceURI(), headerName.getLocalPart(), markAsUnderstood);
    }
    
    @NotNull
    @Deprecated
    public Iterator<Header> getHeaders(@NotNull String nsUri) {
        return this.getHeaders(nsUri, true);
    }
    
    @NotNull
    @Override
    public Iterator<Header> getHeaders(@NotNull final String nsUri, final boolean markAsUnderstood) {
        return new Iterator<Header>() {
            int idx = 0;
            Header next;
            
            @Override
            public boolean hasNext() {
                if (this.next == null) {
                    this.fetch();
                }
                return this.next != null;
            }
            
            @Override
            public Header next() {
                if (this.next == null) {
                    this.fetch();
                    if (this.next == null) {
                        throw new NoSuchElementException();
                    }
                }
                if (markAsUnderstood) {
                    assert HeaderList.this.get(this.idx - 1) == this.next;
                    HeaderList.this.understood(this.idx - 1);
                }
                Header r = this.next;
                this.next = null;
                return r;
            }
            
            private void fetch() {
                while (this.idx < HeaderList.this.size()) {
                    Header h = HeaderList.this.get(this.idx++);
                    if (h.getNamespaceURI().equals(nsUri)) {
                        this.next = h;
                        break;
                    }
                    continue;
                }
            }
            
            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }
    
    public String getTo(AddressingVersion av, SOAPVersion sv) {
        return AddressingUtils.getTo(this, av, sv);
    }
    
    public String getAction(@NotNull AddressingVersion av, @NotNull SOAPVersion sv) {
        return AddressingUtils.getAction(this, av, sv);
    }
    
    public WSEndpointReference getReplyTo(@NotNull AddressingVersion av, @NotNull SOAPVersion sv) {
        return AddressingUtils.getReplyTo(this, av, sv);
    }
    
    public WSEndpointReference getFaultTo(@NotNull AddressingVersion av, @NotNull SOAPVersion sv) {
        return AddressingUtils.getFaultTo(this, av, sv);
    }
    
    public String getMessageID(@NotNull AddressingVersion av, @NotNull SOAPVersion sv) {
        return AddressingUtils.getMessageID(this, av, sv);
    }
    
    public String getRelatesTo(@NotNull AddressingVersion av, @NotNull SOAPVersion sv) {
        return AddressingUtils.getRelatesTo(this, av, sv);
    }
    
    public void fillRequestAddressingHeaders(Packet packet, AddressingVersion av, SOAPVersion sv, boolean oneway, String action, boolean mustUnderstand) {
        AddressingUtils.fillRequestAddressingHeaders(this, packet, av, sv, oneway, action, mustUnderstand);
    }
    
    public void fillRequestAddressingHeaders(Packet packet, AddressingVersion av, SOAPVersion sv, boolean oneway, String action) {
        AddressingUtils.fillRequestAddressingHeaders(this, packet, av, sv, oneway, action);
    }
    
    public void fillRequestAddressingHeaders(WSDLPort wsdlPort, @NotNull WSBinding binding, Packet packet) {
        AddressingUtils.fillRequestAddressingHeaders(this, wsdlPort, binding, packet);
    }
    
    @Override
    public boolean add(Header header) {
        return super.add(header);
    }
    
    @Nullable
    @Override
    public Header remove(@NotNull String nsUri, @NotNull String localName) {
        int len = this.size();
        int i = 0;
        while (i < len) {
            Header h = this.get(i);
            if (h.getLocalPart().equals(localName) && h.getNamespaceURI().equals(nsUri)) {
                return this.remove(i);
            }
            ++i;
        }
        return null;
    }
    
    @Override
    public boolean addOrReplace(Header header) {
        int i = 0;
        while (i < this.size()) {
            Header hdr = this.get(i);
            if (hdr.getNamespaceURI().equals(header.getNamespaceURI()) && hdr.getLocalPart().equals(header.getLocalPart())) {
                this.removeInternal(i);
                this.addInternal(i, header);
                return true;
            }
            ++i;
        }
        return this.add(header);
    }
    
    @Override
    public void replace(Header old, Header header) {
        int i = 0;
        while (i < this.size()) {
            Header hdr = this.get(i);
            if (hdr.getNamespaceURI().equals(header.getNamespaceURI()) && hdr.getLocalPart().equals(header.getLocalPart())) {
                this.removeInternal(i);
                this.addInternal(i, header);
                return;
            }
            ++i;
        }
        throw new IllegalArgumentException();
    }
    
    protected void addInternal(int index, Header header) {
        super.add(index, header);
    }
    
    protected Header removeInternal(int index) {
        return (Header)super.remove(index);
    }
    
    @Nullable
    @Override
    public Header remove(@NotNull QName name) {
        return this.remove(name.getNamespaceURI(), name.getLocalPart());
    }
    
    @Override
    public Header remove(int index) {
        this.removeUnderstoodBit(index);
        return (Header)super.remove(index);
    }
    
    private void removeUnderstoodBit(int index) {
        assert index < this.size();
        if (index < 32) {
            int shiftedUpperBits = this.understoodBits >>> -31 + index << index;
            int lowerBits = this.understoodBits << -index >>> 31 - index >>> 1;
            this.understoodBits = (shiftedUpperBits | lowerBits);
            if (this.moreUnderstoodBits != null && this.moreUnderstoodBits.cardinality() > 0) {
                if (this.moreUnderstoodBits.get(0)) {
                    this.understoodBits |= Integer.MIN_VALUE;
                }
                this.moreUnderstoodBits.clear(0);
                for (int j = this.moreUnderstoodBits.nextSetBit(1); j > 0; j = this.moreUnderstoodBits.nextSetBit(j + 1)) {
                    this.moreUnderstoodBits.set(j - 1);
                    this.moreUnderstoodBits.clear(j);
                }
            }
        }
        else if (this.moreUnderstoodBits != null && this.moreUnderstoodBits.cardinality() > 0) {
            index -= 32;
            this.moreUnderstoodBits.clear(index);
            for (int i = this.moreUnderstoodBits.nextSetBit(index); i >= 1; i = this.moreUnderstoodBits.nextSetBit(i + 1)) {
                this.moreUnderstoodBits.set(i - 1);
                this.moreUnderstoodBits.clear(i);
            }
        }
        if (this.size() - 1 <= 33 && this.moreUnderstoodBits != null) {
            this.moreUnderstoodBits = null;
        }
        return;
    }
    
    @Override
    public boolean remove(Object o) {
        if (o != null) {
            int index = 0;
            while (index < this.size()) {
                if (o.equals(this.get(index))) {
                    this.remove(index);
                    return true;
                }
                ++index;
            }
        }
        return false;
    }
    
    public Header remove(Header h) {
        if (this.remove(h)) {
            return h;
        }
        return null;
    }
    
    public static HeaderList copy(MessageHeaders original) {
        if (original == null) {
            return null;
        }
        return new HeaderList(original);
    }
    
    public static HeaderList copy(HeaderList original) {
        return copy(original);
    }
    
    public void readResponseAddressingHeaders(WSDLPort wsdlPort, WSBinding binding) {
    }
    
    @Override
    public void understood(QName name) {
        this.get(name, true);
    }
    
    @Override
    public void understood(String nsUri, String localName) {
        this.get(nsUri, localName, true);
    }
    
    @Override
    public Set<QName> getUnderstoodHeaders() {
        Set<QName> understoodHdrs = new HashSet();
        for (int i = 0; i < this.size(); ++i) {
            if (this.isUnderstood(i)) {
                Header header = this.get(i);
                understoodHdrs.add(new QName(header.getNamespaceURI(), header.getLocalPart()));
            }
        }
        return understoodHdrs;
    }
    
    @Override
    public boolean isUnderstood(Header header) {
        return this.isUnderstood(header.getNamespaceURI(), header.getLocalPart());
    }
    
    @Override
    public boolean isUnderstood(String nsUri, String localName) {
        int i = 0;
        while (i < this.size()) {
            Header h = this.get(i);
            if (h.getLocalPart().equals(localName) && h.getNamespaceURI().equals(nsUri)) {
                return this.isUnderstood(i);
            }
            ++i;
        }
        return false;
    }
    
    @Override
    public boolean isUnderstood(QName name) {
        return this.isUnderstood(name.getNamespaceURI(), name.getLocalPart());
    }
    
    @Override
    public Set<QName> getNotUnderstoodHeaders(Set<String> roles, Set<QName> knownHeaders, WSBinding binding) {
        Set<QName> notUnderstoodHeaders = null;
        if (roles == null) {
            roles = new HashSet();
        }
        SOAPVersion effectiveSoapVersion = this.getEffectiveSOAPVersion(binding);
        roles.add(effectiveSoapVersion.implicitRole);
        for (int i = 0; i < this.size(); ++i) {
            if (!this.isUnderstood(i)) {
                Header header = this.get(i);
                if (!header.isIgnorable(effectiveSoapVersion, roles)) {
                    QName qName = new QName(header.getNamespaceURI(), header.getLocalPart());
                    if (binding == null) {
                        if (notUnderstoodHeaders == null) {
                            notUnderstoodHeaders = new HashSet();
                        }
                        notUnderstoodHeaders.add(qName);
                    }
                    else if (binding instanceof SOAPBindingImpl && !((SOAPBindingImpl)binding).understandsHeader(qName) && !knownHeaders.contains(qName)) {
                        if (notUnderstoodHeaders == null) {
                            notUnderstoodHeaders = new HashSet();
                        }
                        notUnderstoodHeaders.add(qName);
                    }
                }
            }
        }
        return notUnderstoodHeaders;
    }
    
    private SOAPVersion getEffectiveSOAPVersion(WSBinding binding) {
        SOAPVersion mySOAPVersion = (this.soapVersion != null) ? this.soapVersion : binding.getSOAPVersion();
        if (mySOAPVersion == null) {
            mySOAPVersion = SOAPVersion.SOAP_11;
            return mySOAPVersion;
        }
        return mySOAPVersion;
    }
    
    public void setSoapVersion(SOAPVersion soapVersion) {
        this.soapVersion = soapVersion;
    }
    
    @Override
    public Iterator<Header> getHeaders() {
        return this.iterator();
    }
    
    @Override
    public List<Header> asList() {
        return this;
    }
}
